import time

def execute(func, *args, **kwargs):
    start_time = time.time()

    print(f"Executing: {func.__name__}")
    result = func(*args, **kwargs)  

    end_time = time.time()

    print(f"Time Spent: {(end_time - start_time):2.5f}")
    return result


import os
import json
import pickle

class JsonDBManager:

    def __init__(self):
        self.jsondb_folder_path = os.path.expanduser("~/Documents/JsonDB")
        self.jsondb_sysinfo_folder_path = os.path.join(self.jsondb_folder_path, "__sysinfo__")

        if not os.path.exists(self.jsondb_folder_path):
            os.makedirs(self.jsondb_folder_path)

        if not os.path.exists(self.jsondb_sysinfo_folder_path):
            os.makedirs(self.jsondb_sysinfo_folder_path)


    def create(self, db_name: str, partition_keys: list[str], sort_keys: list[str]) -> bool:
        """Creates a new database. Initializes the database

        :param db_name str: Database name
        :param partition_keys list[str]: List of partition key names
        :param sort_keys list[str]: List of sort key names

        :return bool: Flag indicating if the creation was successful
        """

        db_sysinfo_file_path = os.path.join(
            self.jsondb_sysinfo_folder_path,
            f"{db_name}.pickle"
        )

        db_folder_path = os.path.join(
            self.jsondb_folder_path,
            db_name
        )

        if os.path.exists(db_sysinfo_file_path):
            return False
        
        if os.path.exists(db_folder_path):
            return False
        
        os.makedirs(db_folder_path)

        self.__save_db_status(
            db_name,
            {
                "partition_keys": partition_keys,
                "sort_keys": sort_keys,
                "db_partitions": []
            }
        )

        return True

    def delete(self, db_name: str) -> bool:
        """Deletes a database
        WARNING: All database data will be lost
        
        :param db_name str: Name of the database to delete
        :return bool: Flag indicating if the deletion process was succeeded
        """

        db_sysinfo_file_path = os.path.join(self.jsondb_sysinfo_folder_path, f"{db_name}.pickle")
        db_folder_path = os.path.join(self.jsondb_folder_path, db_name)

        if not os.path.exists(db_folder_path):
            return False
        
        if not os.path.exists(db_sysinfo_file_path):
            return False
        
        for db_file_name in os.listdir(db_folder_path):
            os.remove(os.path.join(db_folder_path, db_file_name))
        
        os.remove(db_sysinfo_file_path)
        os.removedirs(db_folder_path)

        return True


    def load_partition_data(self, db_name: str, partition_name: str) -> list:
        """Load a partition file and returns the items
        
        :param db_name str: Name of the database
        :param partition_name str: Name of the partition file

        :return list: partition items
        """

        partition_file_path = f"{self.jsondb_folder_path}/{db_name}/{partition_name}.json"

        if not os.path.exists(partition_file_path):
            return []
        
        with open(partition_file_path, "r+") as partition_file:
            partition_data = partition_file.read()
        
        return json.loads(partition_data)
    

    def write_partition_data(self, db_name: str, partition_name: str, data: dict) -> bool:
        """Writes data into a database partition

        :param db_name str: Name of the database
        :param partition_name str: Name of the partition file
        :param data dict: Data to write

        :return bool: Flag indicating if the writting process was succeeded
        """

        if not isinstance(data, dict):
            return False

        partition_file_path = f"{self.jsondb_folder_path}/{db_name}/{partition_name}.json"

        status = self.__load_db_status(db_name)
        if not os.path.exists(partition_file_path) and not partition_name in status["db_partitions"]:
            status["db_partitions"].append(partition_name)
            self.__save_db_status(db_name, status)

        with open(partition_file_path, "w+") as partition_file:
            partition_file.write(json.dumps(data, indent=4))
        
        return True
    
    
    def delete_partition(self, db_name: str, partition_name: str) -> bool:
        """Delete a partition
        WARNING: All the partition data will be lost

        :param db_name str: Name of the database
        :param partition_name str: Name of the file partition
        
        :return bool: Flag indicating if the process was succeeded
        """

        partition_file_path = f"{self.jsondb_folder_path}/{db_name}/{partition_name}.json"

        if not os.path.exists(partition_file_path): 
            return False

        status = self.__load_db_status(db_name)
        status["db_partitions"].remove(partition_name)
        self.__save_db_status(db_name, status)

        os.remove(partition_file_path)

        return True


    def list_databases(self) -> list:
        """List all the created databases

        :return list: List of databases
        """

        db_list = os.listdir(self.jsondb_folder_path)
        db_list.remove("__sysinfo__")

        return db_list
    

    def list_database_partitions(self, db_name: str) -> list:
        """List partitions within a database

        :param db_name str: Name of the database
        
        :return list: List of database partitions
        """

        if not self.database_exists(db_name): 
            return []

        db_status = self.__load_db_status(db_name)

        return db_status["db_partitions"]
    

    def database_exists(self, db_name: str) -> bool:
        """Checks if a database exists
        
        :param db_name str: Name of the database
        
        :return bool: Flag indicating if the database exists
        """

        db_file_path = os.path.join(self.jsondb_sysinfo_folder_path, f"{db_name}.pickle")

        if not os.path.exists(db_file_path):
            return False
        
        return True


    def database_partition_exists(self, db_name: str, partition_filename: str) -> bool:
        """Checks if a partition exists within a database

        :param db_name str: Name of the database
        :param partition_filename str: Name of the partition

        :return bool: Flag indicating if the database exists
        """
        partition_file_path = os.path.join(
            self.jsondb_folder_path,
            db_name,
            f"{partition_filename}.json"
        )

        return os.path.exists(partition_file_path)


    def get_database_partition_keys(self, db_name: str) -> list:
        """Returns database partition key list

        :param db_name str: Name of the database
        
        :return list: List containing the partition keys
        """

        if not self.database_exists(db_name):
            return []
        
        db_status = self.__load_db_status(db_name)

        return db_status["partition_keys"]
    

    def get_database_sort_keys(self, db_name: str) -> list:
        """Returns database sort key list

        :param db_name str: Name of the database
        
        :return list: List containing the sort keys
        """

        if not self.database_exists(db_name):
            return []

        db_status = self.__load_db_status(db_name)

        return db_status["sort_keys"]
        

    def __db_folder_exists(self, db_name):
        """Checks if a database folder exists (if the database was initialized)
        
        :param db_name str: Name of the database
        
        :return bool: Flag indicating if the database folder exists
        """

        db_folder_path = os.path.join(
            self.jsondb_folder_path,
            db_name
        )

        return os.path.exists(db_folder_path)
    

    def __db_status_file_exists(self, db_name: str) -> bool:
        """Check if a database status file exists (if the database was initialized)

        :param db_name str: Name of the database
        
        :return bool: Flag indicating if the database exists
        """
        
        db_status_file_path = os.path.join(
            self.jsondb_sysinfo_folder_path,
            f"{db_name}.pickle"
        )

        return os.path.exists(db_status_file_path)


    def __load_db_status(self, db_name: str) -> dict:
        """Load database status information

        :param db_name str: Name of the database

        :return dict: Database status dict
        """

        if not self.__db_status_file_exists(db_name):
            return {}
        
        db_status_file_path = os.path.join(
            self.jsondb_sysinfo_folder_path,
            f"{db_name}.pickle"
        )

        with open(db_status_file_path, "rb") as db_status_pickled:
            db_status = pickle.load(db_status_pickled)

        return db_status
    

    def __save_db_status(self, db_name: str, db_status: dict) -> bool: 
        """Overwrite status of database

        :param db_name str: Name of the database
        :param db_status dict: New database status

        :return bool: Flag indicating if the status update was succesful
        """


        db_status_file_path = os.path.join(
            self.jsondb_sysinfo_folder_path,
            f"{db_name}.pickle"
        )

        with open(db_status_file_path, "wb") as db_status_pickled:
            pickle.dump(db_status, db_status_pickled)

        return True 
        
        


from flask import Flask, request
from JPyDB.JsonDB import JsonDB
from JPyDB.JsonDBManager import JsonDBManager
import argparse
import sys

app = Flask(__name__)
json_db_client = JsonDB()
json_db_manager = JsonDBManager()


@app.route("/get_items", methods=["GET"])
def get_items():
    db_name = request.args.get("db_name")
    
    return json_db_client.get_items(db_name)


@app.route("/put_items", methods=["PUT"])
def put_items():
    db_name = request.args.get("db_name")
    items = request.get_json()
    
    return str(json_db_client.put_items(db_name, items))


@app.route("/query_items", methods=["POST"])
def query_items():
    db_name = request.args.get("db_name")
    query_list = request.get_json()

    return json_db_client.query_items(db_name, query_list)


@app.route("/delete_items", methods=["POST"])
def delete_items():
    db_name = request.args.get("db_name")
    query_list = request.get_json()

    return str(json_db_client.delete_items(db_name, query_list))


@app.route("/create", methods=["POST"])
def create():
    db_name = request.args.get("db_name")
    db_keys = request.get_json()

    return str(json_db_manager.create(db_name, **db_keys))


@app.route("/delete", methods=["GET"])
def delete():
    db_name = request.args.get("db_name")

    return str(json_db_manager.delete(db_name))


@app.route("/list_databases", methods=["GET"])
def get_databases():
    return json_db_manager.list_databases()

def main():

    if len(sys.argv) < 1:
        print("No port specified")
        sys.exit(0)


    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--port",
        "-p",
        type = int,
        default = 10000,
        help = "Port number by default port 10000 will be used"
    )

    args = parser.parse_args().__dict__

    app.run(
        host="localhost",
        port=args["port"]
    )


import os
from JPyDB.JsonDBManager import JsonDBManager


class JsonDB:

    def __init__(self): 
        self.db_manager = JsonDBManager()


    def get_items(self, db_name): 
        """Returns all database items

        :param db_name str: Name of the database

        :return list[dict]: List of database items
        """ 
        
        items = []

        for partition_name in self.db_manager.list_database_partitions(db_name):
            items = [*items, *self.db_manager.load_partition_data(db_name, partition_name).values()]

        return items
    

    def put_items(self, db_name, items):
        """Upserts items in the database

        :param db_name str: Name of the database
        :param items list|dict: List of items or item to insert
        
        :return bool: Flag indicating if the process was succeeded
        """

        partition_keys = self.db_manager.get_database_partition_keys(db_name)
        sort_keys = self.db_manager.get_database_sort_keys(db_name)

        if isinstance(items, dict):

            partition_filename = self.__get_partition_filename(partition_keys, items)
            sort_index = self.__get_sort_index(sort_keys, items)
            partition_data = {sort_index: items}
            if self.db_manager.database_partition_exists(db_name, partition_filename):
                
                partition_data = self.db_manager.load_partition_data(db_name, partition_filename)
                partition_data[sort_index] = items

            self.db_manager.write_partition_data(db_name, partition_filename, partition_data)

        elif isinstance(items, list):

            for item in items:
                partition_filename = self.__get_partition_filename(partition_keys, item)
                sort_index = self.__get_sort_index(sort_keys, item)

            
                partition_data = {sort_index: item}
                if self.db_manager.database_partition_exists(db_name, partition_filename):

                    partition_data = self.db_manager.load_partition_data(db_name, partition_filename)
                    partition_data[sort_index] = item

                self.db_manager.write_partition_data(db_name, partition_filename, partition_data)

        return True


    def query_items(self, db_name, query_list):
        """Search for similar items into a database

        :param db_name str: Name of the database
        :param query_list list[dict]: List of items to search 
        
        :return list: List of found items 
        """

        items = []

        db_partition_keys = self.db_manager.get_database_partition_keys(db_name)
        db_sort_keys = self.db_manager.get_database_sort_keys(db_name)

        for query in query_list:

            partition_filename = self.__get_partition_filename(
                db_partition_keys,
                query
            )

            sort_key_name = self.__get_sort_index(
                db_sort_keys,
                query
            )

            if partition_filename:
                partition_data = self.db_manager.load_partition_data(db_name, partition_filename)
                if not partition_data: continue

                if sort_key_name and sort_key_name in partition_data:
                    partition_item = partition_data[sort_key_name]
                    is_valid = True

                    for attr_key, attr_value in query.items():
                        if attr_key not in partition_item: is_valid = False; break
                        if partition_item[attr_key] != attr_value: is_valid = False; break

                    if is_valid: items.append(partition_item)

                elif not sort_key_name:
                    for partition_item in partition_data.values():
                        is_valid = True

                        for attr_key, attr_value in query.items():
                            if attr_key not in partition_item: is_valid = False; break
                            if partition_item[attr_key] != attr_value: is_valid = False; break

                        if is_valid: items.append(partition_item)

            else:
                for partition_name in self.db_manager.list_database_partitions(db_name):
                    partition_data = self.db_manager.load_partition_data(db_name, partition_name)
                    if not partition_data: continue

                    if sort_key_name and sort_key_name in partition_data:
                        partition_item = partition_data[sort_key_name]
                        is_valid = True

                        for attr_key, attr_value in query.items():
                            if attr_key not in partition_item: is_valid = False; break
                            if partition_item[attr_key] != attr_value: is_valid = False; break

                        if is_valid: items.append(partition_item)

                    elif not sort_key_name:
                        for partition_item in partition_data.values():
                            is_valid = True

                            for attr_key, attr_value in query.items():
                                if attr_key not in partition_item: is_valid = False; break
                                if partition_item[attr_key] != attr_value: is_valid = False; break

                            if is_valid: items.append(partition_item)
        
        return items
    

    def delete_items(self, db_name, query_list):
        """Deletes similar items into a database 

        :param db_name str: Name of the database
        :param query_list list[dict]: List of items to delete

        :return bool: Flag indicating if the process was succeeded
        """

        sort_keys_to_delete = []

        db_partition_keys = self.db_manager.get_database_partition_keys(db_name)
        db_sort_keys = self.db_manager.get_database_sort_keys(db_name)

        for query in query_list:

            partition_filename = self.__get_partition_filename(
                db_partition_keys,
                query
            )

            sort_key_name = self.__get_sort_index(
                db_sort_keys,
                query
            )


            if partition_filename:
                partition_data = self.db_manager.load_partition_data(db_name, partition_filename)
                if not partition_data: continue

                if sort_key_name and sort_key_name in partition_data:
                    partition_item = partition_data[sort_key_name]
                    is_valid = True

                    for attr_key, attr_value in query.items():
                        if attr_key not in partition_item: is_valid = False; break
                        if partition_item[attr_key] != attr_value: is_valid = False; break

                    if is_valid: 
                        del partition_data[sort_key_name]

                        if not partition_data:
                            self.db_manager.delete_partition(db_name, partition_name)

                        else:
                            self.db_manager.write_partition_data(db_name, partition_name, partition_data)


                elif not sort_key_name:

                    sort_keys_to_delete = []

                    for partition_item in partition_data.values():
                        is_valid = True

                        for attr_key, attr_value in query.items():
                            if attr_key not in partition_item: is_valid = False; break
                            if partition_item[attr_key] != attr_value: is_valid = False; break

                        if is_valid: 
                            sort_key_name = ""
                            for sort_key in db_sort_keys:
                                sort_key_name += f"{sort_key}={partition_item[sort_key]}"

                            sort_keys_to_delete.append(sort_key_name)


                    for sort_key_name in sort_keys_to_delete:
                        del partition_data[sort_key_name]
                    
                        if not partition_data:
                            self.db_manager.delete_partition(db_name, partition_name)

                        else:
                            self.db_manager.write_partition_data(db_name, partition_name, partition_data)


            else:
                for partition_name in self.db_manager.list_database_partitions(db_name):
                    partition_data = self.db_manager.load_partition_data(db_name, partition_name)
                    if not partition_data: continue

                    if sort_key_name and sort_key_name in partition_data:
                        partition_item = partition_data[sort_key_name]
                        is_valid = True

                        for attr_key, attr_value in query.items():
                            if attr_key not in partition_item: is_valid = False; break
                            if partition_item[attr_key] != attr_value: is_valid = False; break

                        if is_valid: 
                            del partition_data[sort_key_name]
                            if not partition_data:
                                self.db_manager.delete_partition(db_name, partition_name)

                            else:
                                self.db_manager.write_partition_data(db_name, partition_name, partition_data)


                    elif not sort_key_name:
                        sort_keys_to_delete = []

                        for partition_item in partition_data.values():
                            is_valid = True

                            for attr_key, attr_value in query.items():
                                if attr_key not in partition_item: is_valid = False; break
                                if partition_item[attr_key] != attr_value: is_valid = False; break

                            if is_valid: 
                                sort_key_name = ""
                                for sort_key in db_sort_keys:
                                    sort_key_name += f"{sort_key}={partition_item[sort_key]}"

                                sort_keys_to_delete.append(sort_key_name)


                        for sort_key_name in sort_keys_to_delete:
                            del partition_data[sort_key_name]
                        
                            if not partition_data:
                                self.db_manager.delete_partition(db_name, partition_name)

                            else:
                                self.db_manager.write_partition_data(db_name, partition_name, partition_data)

        return True        


    def __get_partition_filename(self, partition_keys, item):
        """Internal function to build partition filenames using a partition key sample and an item

        :param partition_keys list[str]: list of database partition key names
        :param item dict: Item used to build the filename
    
        :return str: partition name if the keys are found
        """

        partition_filename = ""

        for partition_key in partition_keys:
            
            if not partition_key in item:
                return ""
            
            partition_filename += f"{partition_key}={item[partition_key]}"

        return partition_filename
    

    def __get_sort_index(self, sort_keys, item):
        """Internal function to build sort key names using a sort key sample and an item

        :param sort_keys list[str]: list of database sort key names
        :param item dict: Item used to build the sort key
    
        :return str: sort key index if all the keys are found
        """
        
        sort_index = ""

        for sort_key in sort_keys:
            
            if not sort_key in item:
                return ""
            
            sort_index += f"{sort_key}={item[sort_key]}"

        return sort_index